/*------------------------------------------------------------------------------
* correlator_e.c: correlator
*
* 2010/10/15 0.1 new
*-----------------------------------------------------------------------------*/
//#define SSE2
#ifdef SSE2
#include <emmintrin.h>
#endif
#include <stdlib.h>

#define PI      3.14159265358979323
#define SCALE   (1.0/16.0)          /* carrier scale (LSB) */

extern void gencode(int sat, double coff, double cs, int ns, short *code);

#ifndef SSE2

/* dot products (2x1,2x2,2x3) ------------------------------------------------*/
static void dot_2_1(const short *a1, const short *a2, const short *b1,
                     int n, double *c1, double *c2)
{
    const short *p1=a1,*p2=a2,*q1=b1;
    
    c1[0]=c2[0]=0.0;
    
    for (;p1<a1+n;p1++,p2++,q1++) {
        c1[0]+=(*p1)*(*q1);
        c2[0]+=(*p2)*(*q1);
    }
}
static void dot_2_2(const short *a1, const short *a2, const short *b1,
                    const short *b2, int n, double *c1, double *c2)
{
    const short *p1=a1,*p2=a2,*q1=b1,*q2=b2;
    
    c1[0]=c1[1]=c2[0]=c2[1]=0.0;
    
    for (;p1<a1+n;p1++,p2++,q1++,q2++) {
        c1[0]+=(*p1)*(*q1);
        c1[1]+=(*p1)*(*q2);
        c2[0]+=(*p2)*(*q1);
        c2[1]+=(*p2)*(*q2);
    }
}
static void dot_2_3(const short *a1, const short *a2, const short *b1,
                    const short *b2, const short *b3, int n, double *c1,
                    double *c2)
{
    const short *p1=a1,*p2=a2,*q1=b1,*q2=b2,*q3=b3;
    
    c1[0]=c1[1]=c1[2]=c2[0]=c2[1]=c2[2]=0.0;
    
    for (;p1<a1+n;p1++,p2++,q1++,q2++,q3++) {
        c1[0]+=(*p1)*(*q1);
        c1[1]+=(*p1)*(*q2);
        c1[2]+=(*p1)*(*q3);
        c2[0]+=(*p2)*(*q1);
        c2[1]+=(*p2)*(*q2);
        c2[2]+=(*p2)*(*q3);
    }
}
/* multiply carrier ----------------------------------------------------------*/
static void mulcarr(int type, double freq, double phi0, double ts, int ns,
                    const char *data, short *I, short *Q)
{
    static short cost[32]={0},sint[32]={0};
    const char *de=data+ns;
    double phi,ps;
    int i,index;
    
    if (cost[0]==0.0) {
        for (i=0;i<32;i++) {
            cost[i]=(short)(cos(PI/16*i)/SCALE);
            sint[i]=(short)(sin(PI/16*i)/SCALE);
        }
    }
    phi=phi0*16/PI+0.5;
    phi-=floor(phi/32)*32; /* initial phase (0<=phi<32) */
    ps=ts*freq*32;         /* phase step */
    
    if (type) { /* complex */
        for (;data<de;data+=2,I++,Q++) {
            index=((int)phi)&0x1F;
            *I=cost[index]*data[0];
            *Q=cost[index]*data[1];
            phi+=ps;
        }
    }
    else { /* real */
        for (;data<de;data++,I++,Q++) {
            index=((int)phi)&0x1F; /* max: 2^31/32/38.2MHz=1.757s */
            *I=cost[index]*data[0];
            *Q=sint[index]*data[0];
            phi+=ps;
        }
    }
}
#else /* SSE2 */

/* multiply and sum ----------------------------------------------------------*/
#define MULSUM(p,q,s) { \
    x=_mm_loadu_si128((__m128i *)(p)); \
    y=_mm_loadu_si128((__m128i *)(q)); \
    y=_mm_madd_epi16(y,x); \
    s=_mm_add_epi32(s,y); \
}
/* sum sums ------------------------------------------------------------------*/
#define SUMSUM(s,r) { \
    _mm_storeu_si128((void *)sum,s); \
    r=sum[0]+sum[1]+sum[2]+sum[3]; \
}
/* dot products (2x1,2x2,2x3) ------------------------------------------------*/
static void dot_2_1(const short *a1, const short *a2, const short *b1,
                     int n, double *c1, double *c2)
{
    const short *p1=a1,*p2=a2,*q1=b1;
    int sum[4];
    __m128i s1,s2,x,y;
    
    s1=_mm_setzero_si128();
    s2=_mm_setzero_si128();
    
    for (;p1<a1+n;p1+=8,p2+=8,q1+=8) {
        MULSUM(p1,q1,s1);
        MULSUM(p2,q1,s2);
    }
    SUMSUM(s1,c1[0]);
    SUMSUM(s2,c2[0]);
}
static void dot_2_2(const short *a1, const short *a2, const short *b1,
                    const short *b2, int n, double *c1, double *c2)
{
    const short *p1=a1,*p2=a2,*q1=b1,*q2=b2;
    int sum[4];
    __m128i s1,s2,s3,s4,x,y;
    
    s1=_mm_setzero_si128();
    s2=_mm_setzero_si128();
    s3=_mm_setzero_si128();
    s4=_mm_setzero_si128();
    
    for (;p1<a1+n;p1+=8,p2+=8,q1+=8,q2+=8) {
        MULSUM(p1,q1,s1);
        MULSUM(p1,q2,s2);
        MULSUM(p2,q1,s3);
        MULSUM(p2,q2,s4);
    }
    SUMSUM(s1,c1[0]);
    SUMSUM(s2,c1[1]);
    SUMSUM(s3,c2[0]);
    SUMSUM(s4,c2[1]);
}
static void dot_2_3(const short *a1, const short *a2, const short *b1,
                    const short *b2, const short *b3, int n, double *c1,
                    double *c2)
{
    const short *p1=a1,*p2=a2,*q1=b1,*q2=b2,*q3=b3;
    int sum[4];
    __m128i s1,s2,s3,s4,s5,s6,x,y;
    
    s1=_mm_setzero_si128();
    s2=_mm_setzero_si128();
    s3=_mm_setzero_si128();
    s4=_mm_setzero_si128();
    s5=_mm_setzero_si128();
    s6=_mm_setzero_si128();
    
    for (;p1<a1+n;p1+=8,p2+=8,q1+=8,q2+=8,q3+=8) {
        MULSUM(p1,q1,s1);
        MULSUM(p1,q2,s2);
        MULSUM(p1,q3,s3);
        MULSUM(p2,q1,s4);
        MULSUM(p2,q2,s5);
        MULSUM(p2,q3,s6);
    }
    SUMSUM(s1,c1[0]);
    SUMSUM(s2,c1[1]);
    SUMSUM(s3,c1[2]);
    SUMSUM(s4,c2[0]);
    SUMSUM(s5,c2[1]);
    SUMSUM(s6,c2[2]);
}
/* multiply carrier ----------------------------------------------------------*/
static void mulcarr(int type, double freq, double phi0, double ts, int ns,
                    const char *data, short *I, short *Q)
{
    static short cost[32]={0},sint[32]={0};
    const char *de=data+ns;
    double phi,ps;
    int i,index[4];
    __m128d x1,x2,dx;
    __m128i i1,i2,ind,mask;
    
    if (cost[0]==0.0) {
        for (i=0;i<32;i++) {
            cost[i]=(short)(cos(PI/16*i)/SCALE);
            sint[i]=(short)(sin(PI/16*i)/SCALE);
        }
    }
    phi=phi0*16/PI+0.5;
    phi-=floor(phi/32)*32; /* initial phase (0<=phi<32) */
    ps=ts*freq*32;         /* phase step */
    x1=_mm_set_pd(phi+ps,phi); phi+=ps*2;
    x2=_mm_set_pd(phi+ps,phi);
    dx=_mm_set1_pd(ps*4);
    mask=_mm_set1_epi32(0x1F);
    
    if (type) { /* complex */
        for (;data<de;I+=4,Q+=4,data+=8) {
            
            /* index=(int)x&mask */
            i1=_mm_cvttpd_epi32(x1);
            i2=_mm_cvttpd_epi32(x2);
            ind=_mm_unpacklo_epi32(i1,i2); /* x[0,2,1,3] */
            ind=_mm_and_si128(ind,mask);
            _mm_storeu_si128((void *)index,ind);
            x1=_mm_add_pd(x1,dx);
            x2=_mm_add_pd(x2,dx);
            
            /* I=cos(x)*data, Q=sin(x)*data */
            I[0]=cost[index[0]]*data[0];
            I[1]=cost[index[2]]*data[2];
            I[2]=cost[index[1]]*data[4];
            I[3]=cost[index[3]]*data[6];
            Q[0]=cost[index[0]]*data[1];
            Q[1]=cost[index[2]]*data[3];
            Q[2]=cost[index[1]]*data[5];
            Q[3]=cost[index[3]]*data[7];
        }
    }
    else { /* real */
        for (;data<de;I+=4,Q+=4,data+=4) {
            
            i1=_mm_cvttpd_epi32(x1);
            i2=_mm_cvttpd_epi32(x2);
            ind=_mm_unpacklo_epi32(i1,i2); /* x[0,2,1,3] */
            ind=_mm_and_si128(ind,mask);
            _mm_storeu_si128((void *)index,ind);
            x1=_mm_add_pd(x1,dx);
            x2=_mm_add_pd(x2,dx);
            
            I[0]=cost[index[0]]*data[0];
            I[1]=cost[index[2]]*data[1];
            I[2]=cost[index[1]]*data[2];
            I[3]=cost[index[3]]*data[3];
            Q[0]=sint[index[0]]*data[0];
            Q[1]=sint[index[2]]*data[1];
            Q[2]=sint[index[1]]*data[2];
            Q[3]=sint[index[3]]*data[3];
        }
    }
}
#endif /* SSE2 */

/* correlator ----------------------------------------------------------------*/
extern void correlator(int type, int sat, double crate, double coff,
                       double freq, double phi0, const char *data, double ts,
                       int ns, int s, int nc, double *I, double *Q)
{
    short *Is,*Qs,*code,*codes;
    int i,na=ns+2*s*nc+1;
    
    if (!(Is   =(short *)malloc(sizeof(short)*ns+32))||
        !(Qs   =(short *)malloc(sizeof(short)*ns+32))||
        !(codes=(short *)malloc(sizeof(short)*na+32))) return;
    
    code=codes+s*nc;
    
    /* multiply carrier */
    mulcarr(type,freq,phi0,ts,ns,data,Is,Qs);
    
    /* generate code */
    gencode(sat,coff+s*nc*ts*crate,ts*crate,na,codes);
    
    /* multiply code and integrate */
    if (nc<=0) {
        dot_2_1(Is,Qs,code,ns,I,Q);
    }
    else {
        dot_2_3(Is,Qs,code,code-s,code+s,ns,I,Q);
    }
    for (i=1;i<nc;i++) {
        dot_2_2(Is,Qs,code-(i+1)*s,code+(i+1)*s,ns,I+1+i*2,Q+1+i*2);
    }
    for (i=0;i<1+2*nc;i++) {
        I[i]*=SCALE;
        Q[i]*=SCALE;
    }
    free(Is); free(Qs); free(codes);
}
